@@ -53,7 +53,7 @@ std::string_view demangle(const Symbol<E> &sym) {
 }
 
 template <typename E>
-InputFile<E>::InputFile(Context<E> &ctx, MappedFile<Context<E>> *mf)
+InputFile<E>::InputFile(Context<E> &ctx, MappedFile *mf)
   : mf(mf), filename(mf->name) {
   if (mf->size < sizeof(ElfEhdr<E>))
     Fatal(ctx) << *this << ": file too small";
@@ -98,22 +98,6 @@ ElfShdr<E> *InputFile<E>::find_section(i64 type) {
   return nullptr;
 }
 
-template <typename E>
-void InputFile<E>::clear_symbols() {
-  for (Symbol<E> *sym : get_global_syms()) {
-    if (__atomic_load_n(&sym->file, __ATOMIC_ACQUIRE) == this) {
-      sym->origin = 0;
-      sym->value = -1;
-      sym->sym_idx = -1;
-      sym->ver_idx = VER_NDX_UNSPECIFIED;
-      sym->is_weak = false;
-      sym->is_imported = false;
-      sym->is_exported = false;
-      __atomic_store_n(&sym->file, nullptr, __ATOMIC_RELEASE);
-    }
-  }
-}
-
 // Find the source filename. It should be listed in symtab as STT_FILE.
 template <typename E>
 std::string_view InputFile<E>::get_source_name() const {
@@ -124,7 +108,7 @@ std::string_view InputFile<E>::get_source_name() const {
 }
 
 template <typename E>
-ObjectFile<E>::ObjectFile(Context<E> &ctx, MappedFile<Context<E>> *mf,
+ObjectFile<E>::ObjectFile(Context<E> &ctx, MappedFile *mf,
                           std::string archive_name, bool is_in_lib)
   : InputFile<E>(ctx, mf), archive_name(archive_name), is_in_lib(is_in_lib) {
   this->is_alive = !is_in_lib;
@@ -132,7 +116,7 @@ ObjectFile<E>::ObjectFile(Context<E> &ctx, MappedFile<Context<E>> *mf,
 
 template <typename E>
 ObjectFile<E> *
-ObjectFile<E>::create(Context<E> &ctx, MappedFile<Context<E>> *mf,
+ObjectFile<E>::create(Context<E> &ctx, MappedFile *mf,
                       std::string archive_name, bool is_in_lib) {
   ObjectFile<E> *obj = new ObjectFile<E>(ctx, mf, archive_name, is_in_lib);
   ctx.obj_pool.emplace_back(obj);
@@ -233,6 +217,32 @@ static void read_riscv_attributes(Context<E> &ctx, ObjectFile<E> &file,
   }
 }
 
+template <typename E>
+static bool is_known_section_type(const ElfShdr<E> &shdr) {
+  u32 ty = shdr.sh_type;
+  u32 flags = shdr.sh_flags;
+
+  if (ty == SHT_PROGBITS ||
+      ty == SHT_NOTE ||
+      ty == SHT_NOBITS ||
+      ty == SHT_INIT_ARRAY ||
+      ty == SHT_FINI_ARRAY ||
+      ty == SHT_PREINIT_ARRAY)
+    return true;
+
+  if (SHT_LOUSER <= ty && ty <= SHT_HIUSER && !(flags & SHF_ALLOC))
+    return true;
+  if (SHT_LOOS <= ty && ty <= SHT_HIOS && !(flags & SHF_OS_NONCONFORMING))
+    return true;
+  if (is_x86<E> && ty == SHT_X86_64_UNWIND)
+    return true;
+  if (is_arm32<E> && (ty == SHT_ARM_EXIDX || ty == SHT_ARM_ATTRIBUTES))
+    return true;
+  if (is_riscv<E> && ty == SHT_RISCV_ATTRIBUTES)
+    return true;
+  return false;
+}
+
 template <typename E>
 void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
   // Read sections
@@ -294,7 +304,7 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
       typename decltype(ctx.comdat_groups)::const_accessor acc;
       ctx.comdat_groups.insert(acc, {signature, ComdatGroup()});
       ComdatGroup *group = const_cast<ComdatGroup *>(&acc->second);
-      comdat_groups.push_back({group, (u32)i, entries.subspan(1)});
+      comdat_groups.push_back({group, (i32)i, entries.subspan(1)});
       break;
     }
     case SHT_REL:
@@ -305,6 +315,10 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
     case SHT_NULL:
       break;
     default:
+      if (!is_known_section_type(shdr))
+        Fatal(ctx) << *this << ": " << name << ": unsupported section type: 0x"
+                   << std::hex << (u32)shdr.sh_type;
+
       // .note.GNU-stack section controls executable-ness of the stack
       // area in GNU linkers. We ignore that section because silently
       // making the stack area executable is too dangerous. Tell our
@@ -361,23 +375,22 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
 
       // Save .llvm_addrsig for --icf=safe.
       if (shdr.sh_type == SHT_LLVM_ADDRSIG && !ctx.arg.relocatable) {
-        if (shdr.sh_link != 0) {
+        // sh_link should be the index of the symbol table section.
+        // Tools that mutates the symbol table, such as objcopy or `ld -r`
+        // tend to not preserve sh_link, so we ignore such section.
+        if (shdr.sh_link != 0)
           llvm_addrsig = std::move(this->sections[i]);
-        } else {
-          Warn(ctx) << *this << ": ignoring .llvm_addrsig section without"
-                    << " sh_link; was the file processed by strip or objcopy -r?";
-        }
         continue;
       }
 
       if (shdr.sh_type == SHT_INIT_ARRAY ||
           shdr.sh_type == SHT_FINI_ARRAY ||
           shdr.sh_type == SHT_PREINIT_ARRAY)
-        ctx.has_init_array = true;
+        this->has_init_array = true;
 
       if (name == ".ctors" || name.starts_with(".ctors.") ||
           name == ".dtors" || name.starts_with(".dtors."))
-        ctx.has_ctors = true;
+        this->has_ctors = true;
 
       if (name == ".eh_frame")
         eh_frame_sections.push_back(this->sections[i].get());
@@ -655,13 +668,13 @@ void ObjectFile<E>::sort_relocations(Context<E> &ctx) {
   }
 }
 
-static size_t find_null(std::string_view data, u64 entsize) {
+static size_t find_null(std::string_view data, i64 pos, i64 entsize) {
   if (entsize == 1)
-    return data.find('\0');
+    return data.find('\0', pos);
 
-  for (i64 i = 0; i <= data.size() - entsize; i += entsize)
-    if (data.substr(i, entsize).find_first_not_of('\0') == data.npos)
-      return i;
+  for (; pos <= data.size() - entsize; pos += entsize)
+    if (data.substr(pos, entsize).find_first_not_of('\0') == data.npos)
+      return pos;
 
   return data.npos;
 }
@@ -686,7 +699,7 @@ static size_t find_null(std::string_view data, u64 entsize) {
 template <typename E>
 static std::unique_ptr<MergeableSection<E>>
 split_section(Context<E> &ctx, InputSection<E> &sec) {
-  if (!sec.is_alive || sec.relsec_idx != -1)
+  if (!sec.is_alive || sec.relsec_idx != -1 || sec.sh_size == 0)
     return nullptr;
 
   const ElfShdr<E> &shdr = sec.shdr();
@@ -704,60 +717,53 @@ split_section(Context<E> &ctx, InputSection<E> &sec) {
   if (addralign == 0)
     addralign = 1;
 
-  std::unique_ptr<MergeableSection<E>> rec(new MergeableSection<E>);
-  rec->parent = MergedSection<E>::get_instance(ctx, sec.name(), shdr.sh_type,
-                                               shdr.sh_flags, entsize, addralign);
-  rec->p2align = sec.p2align;
-
-  if (sec.sh_size == 0)
-    return rec;
+  std::unique_ptr<MergeableSection<E>> m(new MergeableSection<E>);
+  m->parent = MergedSection<E>::get_instance(ctx, sec.name(), shdr.sh_type,
+                                             shdr.sh_flags, entsize, addralign);
+  m->p2align = sec.p2align;
 
   // If thes section contents are compressed, uncompress them.
   sec.uncompress(ctx);
 
   std::string_view data = sec.contents;
-  const char *begin = data.data();
-  HyperLogLog estimator;
+  m->contents = sec.contents;
+
+  if (data.size() > UINT32_MAX)
+    Fatal(ctx) << sec << ": mergeable section too large";
 
   // Split sections
   if (shdr.sh_flags & SHF_STRINGS) {
-    while (!data.empty()) {
-      size_t end = find_null(data, entsize);
+    for (i64 pos = 0; pos < data.size();) {
+      m->frag_offsets.push_back(pos);
+      size_t end = find_null(data, pos, entsize);
       if (end == data.npos)
         Fatal(ctx) << sec << ": string is not null terminated";
-
-      std::string_view substr = data.substr(0, end + entsize);
-      data = data.substr(end + entsize);
-
-      rec->strings.push_back(substr);
-      rec->frag_offsets.push_back(substr.data() - begin);
-
-      u64 hash = hash_string(substr);
-      rec->hashes.push_back(hash);
-      estimator.insert(hash);
+      pos = end + entsize;
     }
   } else {
     if (data.size() % entsize)
       Fatal(ctx) << sec << ": section size is not multiple of sh_entsize";
+    m->frag_offsets.reserve(data.size() / entsize);
 
-    while (!data.empty()) {
-      std::string_view substr = data.substr(0, entsize);
-      data = data.substr(entsize);
-
-      rec->strings.push_back(substr);
-      rec->frag_offsets.push_back(substr.data() - begin);
-
-      u64 hash = hash_string(substr);
-      rec->hashes.push_back(hash);
-      estimator.insert(hash);
-    }
+    for (i64 pos = 0; pos < data.size(); pos += entsize)
+      m->frag_offsets.push_back(pos);
   }
 
-  rec->parent->estimator.merge(estimator);
+  // Compute hashes for section pieces
+  HyperLogLog estimator;
+  m->hashes.reserve(m->frag_offsets.size());
+
+  for (i64 i = 0; i < m->frag_offsets.size(); i++) {
+    u64 hash = hash_string(m->get_contents(i));
+    m->hashes.push_back(hash);
+    estimator.insert(hash);
+  }
+
+  m->parent->estimator.merge(estimator);
 
   static Counter counter("string_fragments");
-  counter += rec->fragments.size();
-  return rec;
+  counter += m->frag_offsets.size();
+  return m;
 }
 
 // Usually a section is an atomic unit of inclusion or exclusion.
@@ -819,14 +825,17 @@ template <typename E>
 void ObjectFile<E>::resolve_section_pieces(Context<E> &ctx) {
   for (std::unique_ptr<MergeableSection<E>> &m : mergeable_sections) {
     if (m) {
-      m->fragments.reserve(m->strings.size());
-      for (i64 i = 0; i < m->strings.size(); i++)
-        m->fragments.push_back(m->parent->insert(ctx, m->strings[i], m->hashes[i],
-                                                 m->p2align));
+      m->fragments.reserve(m->frag_offsets.size());
 
-      // Shrink vectors that we will never use again to reclaim memory.
-      m->strings.clear();
+      for (i64 i = 0; i < m->frag_offsets.size(); i++) {
+        SectionFragment<E> *frag =
+          m->parent->insert(ctx, m->get_contents(i), m->hashes[i], m->p2align);
+        m->fragments.push_back(frag);
+      }
+
+      // Reclaim memory as we'll never use this vector again
       m->hashes.clear();
+      m->hashes.shrink_to_fit();
     }
   }
 
@@ -1054,14 +1063,9 @@ ObjectFile<E>::mark_live_objects(Context<E> &ctx,
     if (sym.is_traced)
       print_trace_symbol(ctx, *this, esym, sym);
 
-    if (esym.is_weak())
-      continue;
-
-    if (!sym.file)
-      continue;
-
-    bool keep = esym.is_undef() || (esym.is_common() && !sym.esym().is_common());
-    if (keep && !sym.file->is_alive.test_and_set()) {
+    if (sym.file && !esym.is_weak() &&
+        (esym.is_undef() || (esym.is_common() && !sym.esym().is_common())) &&
+        !sym.file->is_alive.test_and_set()) {
       feeder(sym.file);
 
       if (sym.is_traced)
@@ -1135,23 +1139,19 @@ void ObjectFile<E>::convert_common_symbols(Context<E> &ctx) {
       continue;
     }
 
-    elf_sections2.push_back({});
-    ElfShdr<E> &shdr = elf_sections2.back();
-    memset(&shdr, 0, sizeof(shdr));
-
-    if (sym.get_type() == STT_TLS) {
+    ElfShdr<E> shdr = {};
+    if (sym.get_type() == STT_TLS)
       shdr.sh_flags = SHF_ALLOC | SHF_WRITE | SHF_TLS;
-    } else {
+    else
       shdr.sh_flags = SHF_ALLOC | SHF_WRITE;
-    }
 
     shdr.sh_type = SHT_NOBITS;
     shdr.sh_size = this->elf_syms[i].st_size;
     shdr.sh_addralign = this->elf_syms[i].st_value;
+    elf_sections2.push_back(shdr);
 
     i64 idx = this->elf_sections.size() + elf_sections2.size() - 1;
-    std::unique_ptr<InputSection<E>> isec =
-      std::make_unique<InputSection<E>>(ctx, *this, idx);
+    auto isec = std::make_unique<InputSection<E>>(ctx, *this, idx);
 
     sym.file = this;
     sym.set_input_section(isec.get());
@@ -1159,7 +1159,6 @@ void ObjectFile<E>::convert_common_symbols(Context<E> &ctx) {
     sym.sym_idx = i;
     sym.ver_idx = ctx.default_version;
     sym.is_weak = false;
-
     sections.push_back(std::move(isec));
   }
 }
@@ -1283,15 +1282,14 @@ std::ostream &operator<<(std::ostream &out, const InputFile<E> &file) {
 }
 
 template <typename E>
-SharedFile<E> *
-SharedFile<E>::create(Context<E> &ctx, MappedFile<Context<E>> *mf) {
+SharedFile<E> *SharedFile<E>::create(Context<E> &ctx, MappedFile *mf) {
   SharedFile<E> *obj = new SharedFile(ctx, mf);
   ctx.dso_pool.emplace_back(obj);
   return obj;
 }
 
 template <typename E>
-SharedFile<E>::SharedFile(Context<E> &ctx, MappedFile<Context<E>> *mf)
+SharedFile<E>::SharedFile(Context<E> &ctx, MappedFile *mf)
   : InputFile<E>(ctx, mf) {
   this->is_alive = !ctx.as_needed;
 }
@@ -1390,32 +1388,31 @@ void SharedFile<E>::parse(Context<E> &ctx) {
 // default version of the library) at load-time.
 template <typename E>
 std::vector<std::string_view> SharedFile<E>::read_verdef(Context<E> &ctx) {
-  std::vector<std::string_view> ret(VER_NDX_LAST_RESERVED + 1);
-
   ElfShdr<E> *verdef_sec = this->find_section(SHT_GNU_VERDEF);
   if (!verdef_sec)
-    return ret;
+    return {};
 
   std::string_view verdef = this->get_string(ctx, *verdef_sec);
   std::string_view strtab = this->get_string(ctx, verdef_sec->sh_link);
 
-  ElfVerdef<E> *ver = (ElfVerdef<E> *)verdef.data();
+  std::vector<std::string_view> vec;
+  u8 *ptr = (u8 *)verdef.data();
 
   for (;;) {
+    ElfVerdef<E> *ver = (ElfVerdef<E> *)ptr;
     if (ver->vd_ndx == VER_NDX_UNSPECIFIED)
       Fatal(ctx) << *this << ": symbol version too large";
 
-    if (ret.size() <= ver->vd_ndx)
-      ret.resize(ver->vd_ndx + 1);
+    if (vec.size() <= ver->vd_ndx)
+      vec.resize(ver->vd_ndx + 1);
 
-    ElfVerdaux<E> *aux = (ElfVerdaux<E> *)((u8 *)ver + ver->vd_aux);
-    ret[ver->vd_ndx] = strtab.data() + aux->vda_name;
+    ElfVerdaux<E> *aux = (ElfVerdaux<E> *)(ptr + ver->vd_aux);
+    vec[ver->vd_ndx] = strtab.data() + aux->vda_name;
     if (!ver->vd_next)
       break;
-
-    ver = (ElfVerdef<E> *)((u8 *)ver + ver->vd_next);
+    ptr += ver->vd_next;
   }
-  return ret;
+  return vec;
 }
 
 template <typename E>
@@ -1462,15 +1459,15 @@ SharedFile<E>::mark_live_objects(Context<E> &ctx,
 }
 
 template <typename E>
-std::span<Symbol<E> *> SharedFile<E>::find_aliases(Symbol<E> *sym) {
+std::span<Symbol<E> *> SharedFile<E>::get_symbols_at(Symbol<E> *sym) {
   assert(sym->file == this);
 
-  std::call_once(init_aliases, [&] {
+  std::call_once(init_sorted_syms, [&] {
     for (Symbol<E> *sym : this->symbols)
       if (sym->file == this)
-        aliases.push_back(sym);
+        sorted_syms.push_back(sym);
 
-    tbb::parallel_sort(aliases.begin(), aliases.end(),
+    tbb::parallel_sort(sorted_syms.begin(), sorted_syms.end(),
                        [](Symbol<E> *a, Symbol<E> *b) {
       const ElfSym<E> &x = a->esym();
       const ElfSym<E> &y = b->esym();
@@ -1478,12 +1475,12 @@ std::span<Symbol<E> *> SharedFile<E>::find_aliases(Symbol<E> *sym) {
     });
   });
 
-  auto [begin, end] = std::equal_range(aliases.begin(), aliases.end(), sym,
-                                       [&](Symbol<E> *x, Symbol<E> *y) {
-    return x->esym().st_value < y->esym().st_value;
+  auto [begin, end] = std::equal_range(sorted_syms.begin(), sorted_syms.end(),
+                                       sym, [&](Symbol<E> *a, Symbol<E> *b) {
+    return a->esym().st_value < b->esym().st_value;
   });
 
-  return {&*begin, &*end};
+  return {&*begin, (size_t)(end - begin)};
 }
 
 // Infer an alignment of a DSO symbol. An alignment of a symbol in other
@@ -1542,9 +1539,8 @@ void SharedFile<E>::populate_symtab(Context<E> &ctx) {
   u8 *strtab = ctx.buf + ctx.strtab->shdr.sh_offset;
   i64 strtab_off = this->strtab_offset;
 
-  for (i64 i = 0, j = this->first_global; j < this->elf_syms.size(); i++, j++) {
-    Symbol<E> &sym = *this->symbols[j];
-    if (sym.file != this || !sym.write_to_symtab)
+  for (i64 i = 0; Symbol<E> *sym : this->get_global_syms()) {
+    if (sym->file != this || !sym->write_to_symtab)
       continue;
 
     U32<E> *xindex = nullptr;
@@ -1552,8 +1548,9 @@ void SharedFile<E>::populate_symtab(Context<E> &ctx) {
       xindex = (U32<E> *)(ctx.buf + ctx.symtab_shndx->shdr.sh_offset) +
                this->global_symtab_idx + i;
 
-    *symtab++ = to_output_esym(ctx, sym, strtab_off, xindex);
-    strtab_off += write_string(strtab + strtab_off, sym.name());
+    *symtab++ = to_output_esym(ctx, *sym, strtab_off, xindex);
+    strtab_off += write_string(strtab + strtab_off, sym->name());
+    i++;
   }
 }
 
